"""Storage Module for Ellar"""

__version__ = "0.1.5"

from .module import StorageModule
from .providers import Provider, get_driver
from .schemas import StorageSetup
from .services import StorageService
from .storage import Container, Object, StorageDriver
from .stored_file import StoredFile

__all__ = [
    "StorageModule",
    "StorageService",
    "StorageSetup",
    "StoredFile",
    "Provider",
    "get_driver",
    "Object",
    "Container",
    "StorageDriver",
]



import typing as t

from ellar.common import IModuleSetup, Module
from ellar.core import Config, ModuleSetup
from ellar.core.modules import DynamicModule, ModuleBase
from ellar.di import ProviderConfig

from ellar_storage.controller import StorageController
from ellar_storage.schemas import StorageSetup
from ellar_storage.services import StorageService
from ellar_storage.storage import StorageDriver


class _ContainerOptions(t.TypedDict):
    key: str


class _StorageSetupKey(t.TypedDict):
    driver: t.Type[StorageDriver]
    options: t.Union[_ContainerOptions, t.Dict[str, t.Any]]


@Module()
class StorageModule(ModuleBase, IModuleSetup):
    @classmethod
    def setup(
        cls,
        default: t.Optional[str] = None,
        disable_storage_controller: bool = False,
        **kwargs: _StorageSetupKey,
    ) -> DynamicModule:
        schema = StorageSetup(
            storages=kwargs,  # type:ignore[arg-type]
            default=default,
            disable_storage_controller=disable_storage_controller,
        )
        return DynamicModule(
            cls,
            providers=[
                ProviderConfig(StorageService, use_value=StorageService(schema)),
            ],
            controllers=[]
            if schema.disable_storage_controller
            else [StorageController],
        )

    @classmethod
    def register_setup(cls) -> ModuleSetup:
        return ModuleSetup(cls, inject=[Config], factory=cls.__register_setup_factory)

    @staticmethod
    def __register_setup_factory(
        module: t.Type["StorageModule"], config: Config
    ) -> DynamicModule:
        if config.get("STORAGE_CONFIG") and isinstance(config.STORAGE_CONFIG, dict):
            schema = StorageSetup(**dict(config.STORAGE_CONFIG))
            return DynamicModule(
                module,
                providers=[
                    ProviderConfig(StorageService, use_value=StorageService(schema)),
                ],
                controllers=[]
                if schema.disable_storage_controller
                else [StorageController],
            )
        raise RuntimeError("Could not find `STORAGE_CONFIG` in application config.")



from libcloud.common.types import LibcloudError  # noqa
from libcloud.common.types import InvalidCredsException  # noqa
from libcloud.common.types import ProviderError  # noqa
from libcloud.common.types import MalformedResponseError  # noqa
from libcloud.common.types import LibcloudError  # noqa
from libcloud.storage.types import ContainerError  # noqa
from libcloud.storage.types import ObjectError  # noqa
from libcloud.storage.types import ContainerAlreadyExistsError  # noqa
from libcloud.storage.types import ContainerDoesNotExistError  # noqa
from libcloud.storage.types import ContainerIsNotEmptyError  # noqa
from libcloud.storage.types import ObjectDoesNotExistError  # noqa
from libcloud.storage.types import ObjectHashMismatchError  # noqa
from libcloud.storage.types import InvalidContainerNameError  # noqa
from libcloud.storage.types import ObjectHashMismatchError  # noqa



import contextlib
import os
import typing as t
import uuid

from ellar.common import UploadFile
from ellar.di import injectable
from starlette.concurrency import run_in_threadpool

from ellar_storage.constants import LOCAL_STORAGE_DRIVER_NAME
from ellar_storage.exceptions import (
    ContainerAlreadyExistsError,
    ObjectDoesNotExistError,
)
from ellar_storage.schemas import StorageSetup
from ellar_storage.storage import Container
from ellar_storage.stored_file import StoredFile
from ellar_storage.utils import get_metadata_file_obj


@injectable
class StorageService:
    """
    Manages lib-cloud registered storage drivers for saving, deleting and retrieving files
    """

    __slots__ = ("_storages", "_storage_default")

    def __init__(self, storage_setup: StorageSetup) -> None:
        result = {}

        for storage_name, value in storage_setup.storages.items():
            if value.driver.name == LOCAL_STORAGE_DRIVER_NAME:
                # if its local storage, we need to create the path
                os.makedirs(value.options["key"], 0o777, exist_ok=True)

            driver = value.driver(**value.options)

            with contextlib.suppress(ContainerAlreadyExistsError):
                driver.create_container(container_name=storage_name)

            storage_container = driver.get_container(container_name=storage_name)
            result[storage_name] = storage_container

        self._storages = result
        self._storage_default = t.cast(str, storage_setup.default)

    def get_container(self, name: t.Optional[str] = None) -> Container:
        """
        Gets the container instance associate to the name,
        return default if name isn't provided.
        """
        if name is None:
            container = self._storages.get(self._storage_default)
            assert container is not None
            return container
        if name in self._storages:
            return self._storages[name]
        raise RuntimeError(f"{name} storage has not been added to Storage Config")

    def save(
        self,
        file: UploadFile,
        upload_storage: t.Optional[str] = None,
        extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> StoredFile:
        extra_ = extra or {}
        extra_.setdefault(
            "meta_data",
            {"content_type": file.content_type, "filename": file.filename},
        )
        extra_.setdefault("content_type", file.content_type)

        return self.save_content(
            name=file.filename or str(uuid.uuid4())[10],
            content=file.file,
            upload_storage=upload_storage,
            metadata={"content_type": file.content_type, "filename": file.filename},
            headers=dict(file.headers),
        )

    def save_content(
        self,
        name: str,
        content: t.Optional[t.Iterator[bytes]] = None,
        upload_storage: t.Optional[str] = None,
        metadata: t.Optional[t.Dict[str, t.Any]] = None,
        extra: t.Optional[t.Dict[str, t.Any]] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
        content_path: t.Optional[str] = None,
    ) -> StoredFile:
        """Save file into provided `upload_storage`"""
        if content is None and content_path is None:
            raise ValueError("Either content or content_path must be specified")

        if metadata is not None:
            extra = {
                "meta_data": metadata,
                "content_type": metadata.get(
                    "content_type", "application/octet-stream"
                ),
            }

        container = self.get_container(upload_storage)

        if (
            container.driver.name == LOCAL_STORAGE_DRIVER_NAME
            and extra is not None
            and extra.get("meta_data", None) is not None
        ):
            """
            Libcloud local storage driver doesn't support metadata, so the metadata
            is saved in the same container with the combination of the original name
            and `.metadata.json` as name
            """
            container.upload_object_via_stream(
                iterator=get_metadata_file_obj(extra["meta_data"]),
                object_name=f"{name}.metadata.json",
            )
        if content_path is not None:
            return StoredFile(
                container.upload_object(
                    file_path=content_path,
                    object_name=name,
                    extra=extra,
                    headers=headers,
                )
            )
        assert content is not None
        return StoredFile(
            container.upload_object_via_stream(
                iterator=content, object_name=name, extra=extra, headers=headers
            )
        )

    def __get_storage_from_path(self, path: str) -> t.Tuple[str, str]:
        path_split = path.split("/")
        if len(path_split) == 1:
            upload_storage, file_id = self._storage_default, path_split[0]
        else:
            upload_storage, file_id = path_split
        return upload_storage, file_id

    def get(self, path: str) -> StoredFile:
        """
        Retrieve the file with `provided` path, path is expected to be `storage_name/file_id`.
        """
        upload_storage, file_id = self.__get_storage_from_path(path)
        return StoredFile(self.get_container(upload_storage).get_object(file_id))

    def delete(self, path: str) -> bool:
        """
        Delete the file with `provided` path.

        The path is expected to be `storage_name/file_id`.
        """
        upload_storage, file_id = self.__get_storage_from_path(path)
        obj = self.get_container(upload_storage).get_object(file_id)

        if obj.driver.name == LOCAL_STORAGE_DRIVER_NAME:
            """Try deleting associated metadata file"""
            with contextlib.suppress(ObjectDoesNotExistError):
                obj.container.get_object(f"{obj.name}.metadata.json").delete()

        return obj.delete()

    async def delete_async(self, path: str) -> bool:
        """Async Delete File Operation"""
        return await run_in_threadpool(self.delete, path)

    async def get_async(self, path: str) -> StoredFile:
        """Async Get File Operation"""
        return await run_in_threadpool(self.get, path)

    async def save_async(
        self,
        file: UploadFile,
        upload_storage: t.Optional[str] = None,
    ) -> StoredFile:
        """Async Save File Operation"""
        return await run_in_threadpool(self.save, file, upload_storage=upload_storage)

    async def save_content_async(
        self,
        name: str,
        content: t.Optional[t.Iterator[bytes]] = None,
        upload_storage: t.Optional[str] = None,
        metadata: t.Optional[t.Dict[str, t.Any]] = None,
        extra: t.Optional[t.Dict[str, t.Any]] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
        content_path: t.Optional[str] = None,
    ) -> StoredFile:
        """Async Save Content Operation"""
        return await run_in_threadpool(
            self.save_content,
            name,
            content=content,
            upload_storage=upload_storage,
            metadata=metadata,
            extra=extra,
            headers=headers,
            content_path=content_path,
        )



import json
import typing as t
from tempfile import SpooledTemporaryFile

from ellar_storage.constants import IN_MEMORY_FILESIZE


def get_metadata_file_obj(
    metadata: t.Dict[str, t.Any],
) -> "SpooledTemporaryFile[bytes]":
    f = SpooledTemporaryFile(IN_MEMORY_FILESIZE)
    f.write(json.dumps(metadata).encode())
    f.seek(0)
    return f



import contextlib
import io
import json
import typing as t

from ellar_storage.constants import LOCAL_STORAGE_DRIVER_NAME
from ellar_storage.exceptions import ObjectDoesNotExistError
from ellar_storage.storage import Object


class StoredFile(io.IOBase):
    """Represents a file that has been stored in a database. This class provides
    a file-like interface for reading the file content.
    """

    def __init__(self, obj: Object) -> None:
        if obj.driver.name == LOCAL_STORAGE_DRIVER_NAME:
            """Retrieve metadata from associated metadata file"""
            try:
                metadata_obj = obj.container.get_object(f"{obj.name}.metadata.json")
                with open(metadata_obj.get_cdn_url()) as metadata_file:
                    obj.meta_data = json.load(metadata_file)
            except ObjectDoesNotExistError:  # pragma: no cover
                pass
        self.name = obj.name
        self.size = obj.size
        self.filename = obj.meta_data.get("filename", "unnamed")
        self.content_type = obj.extra.get(
            "content_type",
            obj.meta_data.get("content_type", "application/octet-stream"),
        )
        self.object = obj

    def get_cdn_url(self) -> t.Optional[str]:
        """Retrieves the CDN URL of the file if available."""
        try:
            return self.object.get_cdn_url()
        except NotImplementedError:  # pragma: no cover
            return None

    def read(self, n: int = -1, chunk_size: t.Optional[int] = None) -> bytes:
        """Reads the content of the file.

        Arguments:
            n: The number of bytes to read. If not specified or set to -1,
                it reads the entire content of the file. Defaults to -1.
            chunk_size: The size of the chunks to read at a time.
                If not specified, the default chunk size of the storage provider will be used.

        """
        return next(
            self.range_as_stream(
                0, end_bytes=n if n > 0 else None, chunk_size=chunk_size
            )
        )

    def close(self) -> None:  # pragma: no cover
        pass  # No need to close;

    def seekable(self) -> bool:
        return False  # Seeking is not supported ; pragma: no cover

    def writable(self) -> bool:
        return False  # Writing is not supported ; pragma: no cover

    def readable(self) -> bool:
        return True  # Reading is supported ; pragma: no cover

    def as_stream(self, chunk_size: t.Optional[int] = None) -> t.Iterator[bytes]:
        return self.object.as_stream(chunk_size=chunk_size)

    def range_as_stream(
        self,
        start_bytes: int,
        end_bytes: t.Optional[int] = None,
        chunk_size: t.Optional[int] = None,
    ) -> t.Iterator[bytes]:
        return self.object.range_as_stream(
            start_bytes=start_bytes,
            end_bytes=end_bytes,
            chunk_size=chunk_size,
        )

    def delete(self) -> bool:
        if self.object.driver.name == LOCAL_STORAGE_DRIVER_NAME:
            """Try deleting associated metadata file"""
            with contextlib.suppress(ObjectDoesNotExistError):
                self.object.container.get_object(
                    f"{self.object.name}.metadata.json"
                ).delete()

        return self.object.delete()



import typing as t

from ellar.pydantic import field_validator, model_validator
from pydantic import BaseModel

from ellar_storage.storage import StorageDriver


class _StorageSetupItem(BaseModel):
    driver: t.Type[StorageDriver]
    options: t.Dict[str, t.Any] = {}

    @field_validator("options", mode="before")
    def pre_options_validate(cls, value: t.Dict) -> t.Any:
        if "key" not in value:
            raise ValueError("Driver Options must have a `key` option ")

        return value


class StorageSetup(BaseModel):
    # default storage name that must exist in `storages`
    # as a key if set else it will default to the first entry in `storages`
    default: t.Optional[str] = None
    # storage configurations
    storages: t.Dict[str, _StorageSetupItem]
    # disable StorageController
    disable_storage_controller: bool = False

    @model_validator(mode="before")
    def post_default_validate(cls, values: t.Dict) -> t.Any:
        storages = values.get("storages")
        default = values.get("default")

        if not storages:
            raise ValueError("At least one storage setup is required storages")

        if not default and storages:
            values["default"] = list(storages.keys())[0]

        if default and default not in storages:
            raise ValueError(f"storages must have a '{default}' as key")
        return values



import typing as t

import ellar.common as ecm
from ellar.common import NotFound
from ellar.core import Request
from libcloud.storage.types import ObjectDoesNotExistError
from starlette.responses import RedirectResponse, StreamingResponse

from ellar_storage.services import StorageService


@ecm.Controller(name="storage", include_in_schema=False)
class StorageController:
    def __init__(self, storage_service: StorageService):
        self._storage_service = storage_service

    @ecm.get("/download/{path:path}", name="download", include_in_schema=False)
    @ecm.file()
    def download_file(self, req: Request, path: str) -> t.Any:
        try:
            res = self._storage_service.get(path)

            if res.get_cdn_url() is None:  # pragma: no cover
                return StreamingResponse(
                    res.object.as_stream(),
                    media_type=res.content_type,
                    headers={
                        "Content-Disposition": f"attachment;filename={res.filename}"
                    },
                )

            if res.object.driver.name != "Local Storage":  # pragma: no cover
                return RedirectResponse(res.get_cdn_url())  # type:ignore[arg-type]

            return {
                "path": res.get_cdn_url(),  # since we are using a local storage, this will return a path to the file
                "filename": res.filename,
                "media_type": res.content_type,
            }

        except ObjectDoesNotExistError as obex:
            raise NotFound() from obex



from libcloud.storage.base import Container  # noqa
from libcloud.storage.base import Object  # noqa
from libcloud.storage.base import StorageDriver  # noqa
from libcloud.storage.base import DEFAULT_CONTENT_TYPE  # noqa
from libcloud.storage.base import CHUNK_SIZE  # noqa
